/*
 * Copyright 2011 NextWindow
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of NextWindow
 * not be used in advertising or publicity pertaining to distribution
 * of the software without specific, written prior permission. 
 * NextWindow makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *
 * THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Copyright 2007 Peter Hutterer
 * Copyright 2009 Przemyslaw Firszt
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Red Hat
 * not be used in advertising or publicity pertaining to distribution
 * of the software without specific, written prior permission.  Red
 * Hat makes no representations about the suitability of this software
 * for any purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <linux/input.h>
#include <linux/types.h>

#include <xf86_OSproc.h>

#include <unistd.h>

#include <xf86.h>
#include <xf86Xinput.h>
#include <exevents.h>
#include <xorgVersion.h>
#include <xkbsrv.h>


#include <xserver-properties.h>
/* 1.6 has properties, but no labels */
#ifdef AXIS_LABEL_PROP
#define HAVE_LABELS
#else
#undef HAVE_LABELS
#endif

#include <stdio.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>
#include <xorg-server.h>
#include <xorgVersion.h>
#include <xf86Module.h>
#include <X11/Xatom.h>

#include "nextwindow.h"

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
/* removed from server, purge when dropping support for server 1.10 */
#define XI86_OPEN_ON_INIT 0x01
#define XI86_CONFIGURED 0x02
#endif

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
static InputInfoPtr nwPreInit(InputDriverPtr drv, IDevPtr dev, int flags);
#else
static int nwPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
#endif
static int nwSwitchMode(ClientPtr client, DeviceIntPtr dev,	int mode);
static void nwUnInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags);
static pointer nwPlug(pointer module, pointer options, int *errmaj, int *errmin);
static void nwUnplug(pointer p);
static void nwReadInput(InputInfoPtr pInfo);
static int nwControl(DeviceIntPtr device, int what);
static int nwControlProc(InputInfoPtr pInfo, xDeviceCtl * control);
static int _nw_init_buttons(DeviceIntPtr device);
static int _nw_init_axes(DeviceIntPtr device);

_X_EXPORT InputDriverRec NEXTWINDOW = {
    1,
    "nextwindow",
    NULL,
    nwPreInit,
    nwUnInit,
    NULL,
    0,
};

static XF86ModuleVersionInfo nwVersionRec =
{
    "nextwindow",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XORG_VERSION_CURRENT,
    PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR,
    PACKAGE_VERSION_PATCHLEVEL,
    ABI_CLASS_XINPUT,
    ABI_XINPUT_VERSION,
    MOD_CLASS_XINPUT,
    {0, 0, 0, 0}
};

_X_EXPORT XF86ModuleData nextwindowModuleData =
{
    &nwVersionRec,
    &nwPlug,
    &nwUnplug
};

const char* btn_labels[] = {BTN_LABEL_PROP_BTN_LEFT, BTN_LABEL_PROP_BTN_RIGHT};
const char* abs_axes[] = {AXIS_LABEL_PROP_ABS_X, AXIS_LABEL_PROP_ABS_Y};

static void
nwUnplug(pointer p)
{
};

static pointer
nwPlug(pointer module,
       pointer options,
       int *errmaj,
       int *errmin)
{
    xf86AddInputDriver(&NEXTWINDOW, module, 0);
    return module;
};

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
static int
NewNwPreInit(InputDriverPtr drv,
             InputInfoPtr pInfo,
             int flags);

static InputInfoPtr
nwPreInit(InputDriverPtr drv,
          IDevPtr dev,
          int flags)
{
    InputInfoPtr pInfo;

    if (!(pInfo = xf86AllocateInput(drv, 0)))
        return NULL;

    pInfo->name = xstrdup(dev->identifier);
    pInfo->flags = 0;
    pInfo->type_name = XI_TOUCHSCREEN; /* see XI.h */
    pInfo->conf_idev = dev;

    if (NewNwPreInit(drv, pInfo, flags) == Success)
    {
        return pInfo;
    }

    xf86DeleteInput(pInfo, 0);
    return NULL;
}

static int
NewNwPreInit(InputDriverPtr drv,
             InputInfoPtr pInfo,
             int flags)
#else
static int
nwPreInit(InputDriverPtr drv,
          InputInfoPtr pInfo,
          int flags)
#endif
{
    int ret = Success;
    NextwindowDevicePtr pNW;
    const char *s;

    pNW = calloc(1, sizeof(NextwindowDeviceRec));
    if (!pNW)
    {
        ret = BadAlloc;
        pInfo->private = NULL;
        goto cleanup;;
    }

    pInfo->private = pNW;
    pInfo->type_name = XI_TOUCHSCREEN;
    pInfo->flags = 0;
    pInfo->read_input = nwReadInput; /* new data avl */
    pInfo->switch_mode = nwSwitchMode; /* toggle absolute/relative mode */
    pInfo->device_control = nwControl; /* enable/disable dev */
    pInfo->control_proc = nwControlProc;

    /* populate pInfo->options */
    xf86CollectInputOptions(pInfo, NULL
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
    , pInfo->options
#endif
    );
    /* process generic options */
    xf86ProcessCommonOptions(pInfo, pInfo->options);

    /* process driver specific options */
    pNW->device = xf86SetStrOption(pInfo->options,
                                         "Device",
                                         "/dev/input/nwfermi_event");
    xf86Msg(X_INFO, "%s: Using device %s.\n", pInfo->name, pNW->device);

	s = xf86FindOptionValue (pInfo->options, "ReportingMode");
	if ((s) && (xf86NameCmp (s, "relative") == 0))
        pNW->reporting_mode = Relative;
	else
        pNW->reporting_mode = Absolute;
	pNW->min_x = xf86SetIntOption( pInfo->options, "MinX", 0 );
	pNW->max_x = xf86SetIntOption( pInfo->options, "MaxX", 32767 );
	pNW->min_y = xf86SetIntOption( pInfo->options, "MinY", 0 );
	pNW->max_y = xf86SetIntOption( pInfo->options, "MaxY", 32767 );

    /* Open sockets, init device files, etc. */
    SYSCALL(pInfo->fd = open(pNW->device, O_RDWR | O_NONBLOCK));
    if (pInfo->fd == -1)
    {
        ret = BadImplementation;
        xf86Msg(X_ERROR, "%s: failed to open %s.",
                pInfo->name, pNW->device);
        pInfo->private = NULL;
        goto cleanup;
    }
    /* do more funky stuff */
    close(pInfo->fd);
    pInfo->fd = -1;
    pInfo->flags |= XI86_OPEN_ON_INIT;
    pInfo->flags |= XI86_CONFIGURED;

cleanup:

    if (ret != Success)
    {
         if (pNW != NULL)
             free(pNW);
    }

    return ret;
}

static void nwUnInit(InputDriverPtr drv,
                         InputInfoPtr pInfo,
                         int flags)
{
    NextwindowDevicePtr pNW = pInfo->private;
    if (pNW->device)
    {
        if(pNW->mask) valuator_mask_free(&pNW->mask);
        if(pNW->axes) free(pNW->axes);
        free(pNW->device);
        pNW->device = NULL;
        /* Common error - pInfo->private must be NULL or valid memoy before
         * passing into xf86DeleteInput */
        pInfo->private = NULL;
    }
    xf86DeleteInput(pInfo, 0);
}

static int
nwSwitchMode(	ClientPtr client,
			DeviceIntPtr device,
			int mode )
{
    InputInfoPtr        pInfo = device->public.devicePrivate;
    NextwindowDevicePtr pNW = pInfo->private;

	if ((mode == Absolute) || (mode == Relative))
	{
		pNW->reporting_mode = mode;
		return (Success);
	}
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) == 0
	else if ((mode == SendCoreEvents) || (mode == DontSendCoreEvents))
	{
		xf86XInputSetSendCoreEvents (local, (mode == SendCoreEvents));
		return (Success);
	}
#endif
	else
		return (!Success);
}

static int
_nw_init_touch(DeviceIntPtr device)
{
    InputInfoPtr        pInfo = device->public.devicePrivate;
    // NextwindowDevicePtr pNW = pInfo->private;
    int                 ret = Success;
    const int           num_touches = 2;

    if (!InitTouchClassDeviceStruct(device, num_touches, XIDirectTouch, 2))
    {
        xf86Msg(X_ERROR, "%s: Failed to register touch device.\n", pInfo->name);
        ret = BadAlloc;
    }

    return ret;
}

static int
_nw_init_buttons(DeviceIntPtr device)
{
    InputInfoPtr        pInfo = device->public.devicePrivate;
    NextwindowDevicePtr pNW = pInfo->private;
    CARD8               *map;
    int                 i;
    int                 ret = Success;
    const int           num_buttons = 2;

    map = calloc(num_buttons, sizeof(CARD8));

    pNW->labels = malloc(sizeof(Atom)*num_buttons);

    for (i = 0; i < num_buttons; i++) {
        pNW->labels[i] = XIGetKnownProperty(btn_labels[i]);
        map[i] = i;
    }

    if (!InitButtonClassDeviceStruct(device, num_buttons, pNW->labels, map))
    {
        xf86Msg(X_ERROR, "%s: Failed to register buttons.\n", pInfo->name);
        ret = BadAlloc;
    }

    return ret;
}

static void nwInitAxesLabels(NextwindowDevicePtr pNW, int natoms, Atom *atoms)
{
#ifdef HAVE_LABELS
    Atom atom;
    int axis;
    int labels_len = pNW->num_vals;

    memset(atoms, 0, natoms * sizeof(Atom));

    /* Now fill the ones we know */
    for (axis = 0; axis < labels_len; axis++)
    {
        atom = XIGetKnownProperty(abs_axes[axis]);
        if (!atom) /* Should not happen */
            continue;

        atoms[axis] = atom;
    }
#endif
}


static int
_nw_init_axes(DeviceIntPtr device)
{
    InputInfoPtr        pInfo = device->public.devicePrivate;
    NextwindowDevicePtr pNW = pInfo->private;
    int                 i;
    const int           num_axes = 2;
    Atom                * atoms;

    pNW->num_vals = num_axes;
    atoms = malloc(pNW->num_vals * sizeof(Atom));
    pNW->axes = atoms;
    nwInitAxesLabels(pNW, pNW->num_vals, atoms);
    pNW->mask = valuator_mask_new(num_axes);
    if (!InitValuatorClassDeviceStruct(device,
                num_axes,
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
                atoms,
#endif	
                GetMotionHistorySize(),
                Absolute))
        return BadAlloc;

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 12
    pInfo->dev->valuator->mode = Absolute;
#endif
#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) < 13
    if (!InitAbsoluteClassDeviceStruct(device))
        return BadAlloc;
#endif

    for (i = 0; i < pNW->num_vals; i++)
    {
        if(i==0) {
            xf86InitValuatorAxisStruct(device, i, pNW->axes[i], pNW->min_x, pNW->max_x, 1, 1, 1
    #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
                , Absolute
    #endif
                );
        } else {
            xf86InitValuatorAxisStruct(device, i, pNW->axes[i], pNW->min_y, pNW->max_y, 1, 1, 1
    #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
                , Absolute
    #endif
                );
        }
        xf86InitValuatorDefaults(device, i);
    }
    return Success;
}

static void _nwPtrControl(DeviceIntPtr dev, PtrCtrl* ctrl)
{
}

static int
nwControlProc(InputInfoPtr pInfo,
			 xDeviceCtl * control )
{
	xDeviceAbsCalibCtl *c = (xDeviceAbsCalibCtl *) control;
	NextwindowDevicePtr pNW = pInfo->private;

	pNW->min_x = c->min_x;
	pNW->max_x = c->max_x;
	pNW->min_y = c->min_y;
	pNW->max_y = c->max_y;
	return (Success);
}

static int nwControl(DeviceIntPtr device,
                         int what)
{
    InputInfoPtr        pInfo = device->public.devicePrivate;
    NextwindowDevicePtr pNW = pInfo->private;

    switch(what)
    {
        case DEVICE_INIT:
            _nw_init_buttons(device);
            _nw_init_axes(device);
            _nw_init_touch(device);
            if (!InitPtrFeedbackClassDeviceStruct(device, _nwPtrControl))
            {
                xf86Msg(X_ERROR, "%s: cannot init ptr feedback.\n", pInfo->name);
                return !Success;
            }
            break;

        /* Switch device on.  Establish socket, start event delivery.  */
        case DEVICE_ON:
            xf86Msg(X_INFO, "%s: On.\n", pInfo->name);
            if (device->public.on)
                    break;

            SYSCALL(pInfo->fd = open(pNW->device, O_RDONLY | O_NONBLOCK));
            if (pInfo->fd < 0)
            {
                xf86Msg(X_ERROR, "%s: cannot open device.\n", pInfo->name);
                return BadRequest;
            }

            xf86FlushInput(pInfo->fd);
            xf86AddEnabledDevice(pInfo);
            device->public.on = TRUE;

            break;
       case DEVICE_OFF:
            xf86Msg(X_INFO, "%s: Off.\n", pInfo->name);
            if (!device->public.on)
                break;
            xf86RemoveEnabledDevice(pInfo);
            close(pInfo->fd);
            pInfo->fd = -1;
            device->public.on = FALSE;
            break;
      case DEVICE_CLOSE:
            /* free what we have to free */
            break;
    }
    return Success;
}

void nwProcessEvent(InputInfoPtr pInfo, struct input_event *ev)
{
    int64_t currevt = 0, lastevt = 0;
    int i, evtype;
    NextwindowDevicePtr pNW = pInfo->private;
    struct timeval tv;

    switch (ev->type)
    {
        case EV_REL:
            break;
        case EV_ABS:
            if((ev->code == ABS_MT_TRACKING_ID)) {
                pNW->update_track = TRUE;
                pNW->new_track_val = ev->value;
            } else if(ev->code == ABS_MT_POSITION_X) {
                if(pNW->new_pos[pNW->new_track_val].x != ev->value) { //filter duplicate events
                    pNW->update_x = TRUE;
                    pNW->new_pos[pNW->new_track_val].x = ev->value;
                }
            } else if(ev->code == ABS_MT_POSITION_Y) {
                if(pNW->new_pos[pNW->new_track_val].y != ev->value) { //filter duplicate events
                    pNW->update_y = TRUE;
                    pNW->new_pos[pNW->new_track_val].y = ev->value;
                }
            }
            break;
        case EV_KEY:
            if (ev->code == BTN_LEFT) {
                pNW->update_btn_left = TRUE;
                pNW->new_btn_left_val = ev->value;
            }
            break;
        case EV_SYN:
        {
            if(pNW->update_track) {
                lastevt = pNW->lasttrack[pNW->new_track_val];
                memcpy(&tv, &ev->time, sizeof(struct timeval));
                currevt = tv.tv_sec*10+tv.tv_usec/100000;
                pNW->lasttrack[pNW->new_track_val] = currevt;
            }
            if(pNW->update_track && (pNW->update_x || pNW->update_y))
            {
                int x = pNW->new_pos[pNW->new_track_val].x;
                int y = pNW->new_pos[pNW->new_track_val].y;
                if(currevt-lastevt>1) { //detect finger down
                    // xf86Msg(X_INFO, "%s: New track info touch %d.\n", pInfo->name, pNW->new_track_val);
                    evtype = XI_TouchBegin;
                } else { //detect finger moved
                    evtype = XI_TouchUpdate;
                }
                valuator_mask_set(pNW->mask, 0, x);
                valuator_mask_set(pNW->mask, 1, y);
                xf86PostTouchEvent(pInfo->dev, pNW->new_track_val, evtype, 0, pNW->mask);
            } else if((pNW->update_btn_left)&&(pNW->new_btn_left_val == 0)) {
                //all fingers are up
                for(i = 0; i < 2; i++) {
                    if(pNW->lasttrack[i] != 0) {
                        // xf86Msg(X_INFO, "%s: Final clean track info touch %d.\n", pInfo->name, i);
                        pNW->new_pos[i].x = 0;
                        pNW->new_pos[i].y = 0;
                        pNW->lasttrack[i] = 0;
                        evtype = XI_TouchEnd;
                        valuator_mask_unset(pNW->mask, 0);
                        valuator_mask_unset(pNW->mask, 1);
                        xf86PostTouchEvent(pInfo->dev, i, evtype, 0, pNW->mask);
                    }
                }
            }
            if(pNW->update_track) {
                //Check other pointer events for funger up by 100ms timeout
                for(i = pNW->new_track_val+1; i < 2; i++) {
                    lastevt = pNW->lasttrack[i];
                    if((lastevt!=0)&&(currevt-lastevt>0)) {
                        // xf86Msg(X_INFO, "%s: Clean track info touch %d.\n", pInfo->name, i);
                        evtype = XI_TouchEnd;
                        pNW->new_pos[i].x = 0;
                        pNW->new_pos[i].y = 0;
                        pNW->lasttrack[i] = 0;
                        valuator_mask_unset(pNW->mask, 0);
                        valuator_mask_unset(pNW->mask, 1);
                        xf86PostTouchEvent(pInfo->dev, i, evtype, 0, pNW->mask);
                    }
                }
            }

            pNW->update_track = FALSE;
            pNW->update_x = FALSE;
            pNW->update_y = FALSE;
            pNW->update_btn_left = FALSE;
            break;
        }
    }
}

// magic number to reduce reads
#define NUM_EVENTS 16

static void nwReadInput(InputInfoPtr pInfo)
{
    struct input_event ev[NUM_EVENTS];
    int i, len = sizeof(ev);

    while (len == sizeof(ev))
    {
        len = read(pInfo->fd, &ev, sizeof(ev));

        if (len < 0)
            break;

        if (len % sizeof(ev[0]))
        {
            // X_NONE does not allocate memory
            xf86MsgVerb(X_NONE, 0, "%s: Read error: %s\n", pInfo->name, strerror(errno)); 
            break;
        }

        for (i = 0; i < len / sizeof(ev[0]); i++)
        {
            nwProcessEvent(pInfo, &ev[i]);
        }
    }
}

